#include "fitonoff.hh"

#include <iostream>


double
FitOnOff::viterbi(double c0, double c1, bool traceback) {
    v[0][0]=0;
    v[1][0]=0;
    
    for (size_type i=1;i<=x.size(); ++i) {
	v[0][i] = (x[i-1]-c0)*(x[i-1]-c0) + std::min(v[0][i-1],v[1][i-1]+delta_10);
	v[1][i] = (x[i-1]-c1)*(x[i-1]-c1) + std::min(v[1][i-1],v[0][i-1]+delta_01);
	
	if (traceback) {
	    t[0][i] = (v[0][i-1] > v[1][i-1]+delta_10); // 0 for going to 0, 1 for going to 1
	    t[1][i] = !(v[1][i-1] > v[0][i-1]+delta_01);
	}
    }

    if (traceback) {
	bool state = v[1][x.size()]<v[0][x.size()]; // 1 if v[1] end smaller, 0 otherwise
	
	for (size_type i=x.size() ;i>0; --i) {
	    trace[i]=state;
	    state=t[state][i];
	}
	trace[0]=0;
    }
    
    return std::min(v[0][x.size()],v[1][x.size()]);
}

pf_t
FitOnOff::forward(double c0, double c1) {
    v[0][0]=1;
    v[1][0]=1;
    
    for (size_type i=1;i<=x.size(); ++i) {
	v[0][i] = exp(-beta*(x[i-1]-c0)*(x[i-1]-c0)) * (v[0][i-1] + v[1][i-1]*exp_delta_10);
	v[1][i] = exp(-beta*(x[i-1]-c1)*(x[i-1]-c1)) * (v[1][i-1] + v[0][i-1]*exp_delta_01);
    }

    return v[0][x.size()] + v[1][x.size()];
}

std::pair<double,double>
FitOnOff::optimize(double c0, double c1) {

    double stepwidth=0.005; //length of the gradient vector added in one iteration
    size_type max_steps=200;


    bool converged=false;
    
    std::vector<pf_t> theta;
    theta.resize(x.size()+1);
    
    std::vector<std::vector<std::vector<pf_t> > > dv; // partial derivations dv[x][y][i] of v[x][i] by cy
    dv.resize(2);
    dv[0].resize(2);
    dv[1].resize(2);
    dv[0][0].resize(x.size()+1);
    dv[0][1].resize(x.size()+1);
    dv[1][0].resize(x.size()+1);
    dv[1][1].resize(x.size()+1);
    
    pf_t last2_c0=c1;
    pf_t last2_c1=c0;
    pf_t last_c0=c1;
    pf_t last_c1=c0;
    
    size_type step=0;
    while (!converged && step<max_steps) {
	step++;

	
	// compute thetas for c0,c1
	double viterbi_score=viterbi(c0,c1);

	//std::cout << "c0: "<<c0<<" c1: "<<c1<<" Score: "<<viterbi_score<<std::endl;
	//write_viterbi_path_compact(std::cout);
    	
	pf_t Theta=1;
	theta[0]=1;
	for (size_type i=1;i<=x.size(); ++i) {
	    //theta[i]=1;
	    theta[i]=exp(beta*(v[0][i]+v[1][i]-v[0][i-1]-v[1][i-1])); // exp(-beta*viterbi_score/x.size());
	    Theta *= theta[i];
	}
		
	
	// determine gradient
	
	// initialize
	v[0][0]=1;
	v[1][0]=1;
	
	dv[0][0][0] = 0;
	dv[0][1][0] = 0;
	dv[1][0][0] = 0;
	dv[1][1][0] = 0;

	// recurse
	for (size_type i=1;i<=x.size(); ++i) {
	    pf_t exp_dev0=exp(-beta*(x[i-1]-c0)*(x[i-1]-c0)); // boltzmann weight of square devitation xi to c0
	    pf_t exp_dev1=exp(-beta*(x[i-1]-c1)*(x[i-1]-c1)); // boltzmann weight of square devitation xi to c1
	    
	    v[0][i] = theta[i]*exp_dev0 * (v[0][i-1] + v[1][i-1]*exp_delta_10);
	    v[1][i] = theta[i]*exp_dev1 * (v[1][i-1] + v[0][i-1]*exp_delta_01);
    	    
	    dv[0][0][i] = theta[i]*
		exp_dev0 * ( dv[0][0][i-1] + dv[1][0][i-1]*exp_delta_10
			     +
			     (v[0][i-1]+v[1][i-1]*exp_delta_10)*2*beta*(x[i-1]-c0));
	    
	    dv[0][1][i] = theta[i] * exp_dev0 * ( dv[0][1][i-1] + dv[1][1][i-1]*exp_delta_10);
	    
	    dv[1][0][i] = theta[i] * exp_dev1 * ( dv[1][0][i-1] + dv[0][0][i-1]*exp_delta_01);
	    
	    dv[1][1][i] = theta[i] *
		exp_dev1 * ( dv[1][1][i-1] + dv[0][1][i-1]*exp_delta_01
			     +
			     (v[1][i-1]+v[0][i-1]*exp_delta_01)*2*beta*(x[i-1]-c1));
	}
	
	pf_t Z = v[0][x.size()]+v[1][x.size()];
	
	
	//std::cout << "Step "<<step <<" Z: "<< (Z/Theta) <<" Z*theta: "<<Z<<" theta: "<<Theta<<std::endl;
	
	pf_t d0 = dv[0][0][x.size()] + dv[1][0][x.size()];	
	pf_t d1 = dv[0][1][x.size()] + dv[1][1][x.size()];
	
	
	// normalize vector (d0,d1) in a numerically nice way
	pf_t dlen_by_d0 = sqrt(1+d1/d0*d1/d0); // = dlen/d0
	pf_t dlen_by_d1 = sqrt(d0/d1*d0/d1+1); // = dlen/d1
						     
	d0=(d0>0?1:-1)*stepwidth*1/dlen_by_d0;
	d1=(d1>0?1:-1)*stepwidth*1/dlen_by_d1;
	
	// pf_t dlen=sqrt(d0*d0+d1*d1);	
	// 	d0/=dlen/stepwidth;
	// 	d1/=dlen/stepwidth;
	
	//std::cout << "d0: "<<d0<<" d1: "<<d1<<std::endl;
	
	c0+=d0;
	c1+=d1;
	
	//print_table("v0",v[0]);
	//print_table("v1",v[1]);
	
	// 	print_table("dv00",dv[0][0]);
	// 	print_table("dv01",dv[0][1]);
	// 	print_table("dv10",dv[1][0]);
	// 	print_table("dv11",dv[1][1]);
	
	converged=fabs(last2_c0-c0)<stepwidth/2 && fabs(last2_c1-c1)<stepwidth/2;
	last2_c0=last_c0;
	last2_c1=last_c1;
	last_c0=c0;
	last_c1=c1;
    }
    
    return std::pair<double,double>(c0,c1);
}


void
FitOnOff::write_viterbi_path(std::ostream &out,double c0, double c1) const {
    out.precision(2);
    for (size_type i=1;i<=x.size(); ++i) {
	out << (trace[i]==0?c0:c1)<<std::endl;
    }
    //out<<std::endl;
}

void
FitOnOff::write_viterbi_path_compact(std::ostream &out) const {
    
    for (size_type i=1;i<=x.size(); ++i) {
	if (trace[i-1]!=trace[i]) {
	    if (trace[i]) out << i << "-";
	    else out << (i-1) << " ";
	}
    }
    out<<std::endl;
}



void
FitOnOff::print_table(const std::string &name, const std::vector<bool> &v) const {
    std::cout <<name<<" ";
    for (size_type i=1;i<=x.size(); ++i) {std::cout << v[i] <<"  ";}
    std::cout <<std::endl;
}
void
FitOnOff::print_table(const std::string &name, const std::vector<pf_t> &v) const {
    std::cout.precision(2);
    std::cout <<name<<" ";
    for (size_type i=1;i<=x.size(); ++i) {std::cout << v[i] <<" ";}
    std::cout <<std::endl;
}


void
FitOnOff::print_tables() const 
{
    print_table("v0",v[0]);
    print_table("v1",v[1]);
    print_table("t0",t[0]);
    print_table("t1",t[1]);
    print_table("tr",trace);
}
